require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/activity_state_template/activity_state_template_attributes'
require 'uu/os/activity_state_template/activity_state_template_create'
require 'uu/os/activity_state_template/activity_state_template_get_activity_state_template_list'
require 'uu/os/activity_state_template/activity_state_template_set_attributes'


# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Activity State Template.
    module ActivityStateTemplate
      extend self

      # Service path
      PATH = 'uu/os/ActivityStateTemplate'

      # Creates a new activity state template on the specified activity template.
      # 
      # @param [String, UU::OS::UESURI] activity_template_uri <b>Not nil.</b> URI of the activity template the new activity state template is to be created on.
      # @param [UU::OS::ActivityStateTemplate::ActivityStateTemplateCreate] options Specific options.
      # @option options [String] :name <b>Not nil.</b> Name of the new activity state template.
      # @option options [String] :code Code of the new activity state template. If not specified, the code is generated automatically using the system sequence. The code must be unique in the scope of the given meta artifact, otherwise the create command fails.
      # @option options [String] :description Description of the new activity state template. If not specified, it is left empty.
      # @option options [String, UU::OS::Activity::ActivityStateType] :type <b>Not nil.</b> Type of the new activity state template. All possible types can be found as constants in {UU::OS::Activity::ActivityStateType} class. It is not possible to use the constant SYSTEM from this class.
      # @option options [String, UU::OS::Activity::ActivityStateSubtype] :subtype Subtype of the new activity state template. All possible subtypes can be found as constants in {UU::OS::Activity::ActivityStateSubtype} class. If not specified, the first possible subtype will be used.
      # @option options [Boolean] :entry_state If this parameter is true, this state will be set as the initial state for activities created according to this template. The default value is false.
      # @option options [Boolean] :exit_state If this parameter is true, this state will be considered the final state of activities created according to this template. The default value is false.
      # @option options [String, UU::OS::UESURI] :icon_uri Icon of the new activity state template. If not specified, there will be used a default icon derived from the specified type. Allowed format of icon: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option options [Integer] :order Order of the new activity state template. All activity state templates are suggested to the user in this order when setting the activity state to an activity which is created according to the given activity template. If it is not set, no specific order is specified.
      # @option options [Array] :participants Set of participants who can set the activity state according to this template. If it is set to the value SYSTEM, it cannot contain any other value in the given set. The default value is ['COMPETENT_ROLE', 'EXECUTIVE_ROLE', 'OWNER']. All possible participants can be found as constants in {UU::OS::ActivityStateTemplate::ActivityStateParticipant} class.
      # @option options [Array] :digital_workspace_visibility Set of roles which will get a digital workspace record in their digital workspace in the case when an activity is set to the state which is created according to this activity state template. If it is set to the value NOBODY, it cannot contain any other value in the given set. The default value is ['COMPETENT_ROLE', 'EXECUTIVE_ROLE']. All possible visibilities can be found as constants in {UU::OS::ActivityStateTemplate::ActivityStateVisibility} class.
      # 
      # @return [UU::OS::UESURI] URI of the new activity state template.
      # 
      # @example
      #   # Create an activity state template
      #   UU::OS::ActivityStateTemplate.create('ues:TERRITORY:META_ARTIFACT:ACTIVITY_TEMPLATE',
      #      :name => 'Activity state template name',
      #      :code => 'Activity state template code',
      #      :description => 'Activity state template description',
      #      :type => 'ACTIVE',
      #      :subtype => 'III',
      #      :icon_uri => 'ues:#{system}:#{ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{S10_C02}',
      #      :order => 30,
      #      :entry_state => false,
      #      :exit_state => false,
      #      :participants => ["COMPETENT_ROLE","EXECUTIVE_ROLE","OWNER","OTHERS"],
      #      :digital_workspace_visibility => ["COMPETENT_ROLE","EXECUTIVE_ROLE"])
      def create(activity_template_uri, options)
        svc = UU::OS::REST::RemoteClient.new(ActivityStateTemplate)
        payload = UU::OS::ActivityStateTemplate::ActivityStateTemplateCreate.new(options).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', activity_template_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns attributes of the specified activity state template.
      # 
      # @param [String, UU::OS::UESURI] activity_state_template_uri <b>Not nil.</b> URI of the activity state template whose attributes are to be returned.
      # 
      # @return [UU::OS::ActivityStateTemplate::ActivityStateTemplateAttributes] Object with attributes of the activity state template.
      # 
      # @example
      #   # Get attributes of an activity state template
      #   UU::OS::ActivityStateTemplate.get_attributes('ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE')
      def get_attributes(activity_state_template_uri)
        svc = UU::OS::REST::RemoteClient.new(ActivityStateTemplate)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', activity_state_template_uri)
          return ActivityStateTemplate::ActivityStateTemplateAttributes.new(res)
        end
      end

      # Sets attributes of an activity state template. This command cannot modify the type.
      # 
      # @param [String, UU::OS::UESURI] activity_state_template_uri <b>Not nil.</b> URI of the activity state template whose attributes are to be modified.
      # @param [UU::OS::ActivityStateTemplate::ActivityStateTemplateSetAttributes] options Specific options.
      # @option options [String] :name The new name of the activity state template. If it is not specified, the name is not modified.
      # @option options [String] :code The new code of the activity state template. If it is not specified, the code is not modified. If blank string is set, the code is generated automatically using the system sequence. The code must be unique in the scope of the given meta artifact, otherwise the setAttributes command fails.
      # @option options [String] :description The new description of the activity state template. If it is not specified, the description is not modified.
      # @option options [String, UU::OS::Activity::ActivityStateSubtype] :subtype The new subtype of the activity state template. All possible subtypes can be found as constants in {UU::OS::Activity::ActivityStateSubtype} class. If it is not specified, the subtype is not modified.
      # @option options [Boolean] :entry_state After an activity is created according to an activity template, it will be set into the state marked as the entry state. There is always just one activity state template marked as the entry state on every activity template. By setting true, the new activity state template will become the entry state instead of the one previously specified. The default value is false. If the user tries to set false to an existing entry state, the setAttributes command fails. If it is not specified, the parameter is not modified.
      # @option options [Boolean] :exit_state There is always just one activity state template marked as the exit state on every activity template. By setting true, the new activity state template will become the exit state instead of the one previously specified. The default value is false. If the user tries to set false to an existing exit state, the setAttributes command fails. If it is not specified, the parameter is not modified.
      # @option options [String, UU::OS::UESURI] :icon_uri URI of the new activity state template's icon. If it is not specified, the icon is not modified. If nil URI ('ues:{[-1]}:{[-1]}:') is set, the activity state template's icon is replaced by a default icon derived from the type. Allowed format of icon: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option options [Integer] :order Order of the activity state template. All activity state templates are suggested to the user in this order when setting the activity state to an activity which is created according to the given activity template. If it is not specified, the parameter is not modified.
      # @option options [Array] :participants Set of participants who can set activity state according to this template. If it is set to the value SYSTEM, it cannot contain any other value in the given set. If it is not specified, the parameter is not modified. All possible participants can be found as constants in {UU::OS::ActivityStateTemplate::ActivityStateParticipant} class.
      # @option options [Array] :digital_workspace_visibility Set of roles which will get a digital workspace record in their digital workspace in the case when an activity is set to the state which is created according to this activity state template. If it is set to the value NOBODY, it cannot contain any other value in the given set. If it is not specified, the parameter is not modified. All possible visibilities can be found as constants in {UU::OS::ActivityStateTemplate::ActivityStateVisibility} class.
      # 
      # @return [UU::OS::UESURI] URI of the modified activity state template.
      # 
      # @example
      #   # Set all attributes of an activity state template
      #   UU::OS::ActivityStateTemplate.set_attributes('ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE',
      #      :name => 'Activity state template name',
      #      :code => 'Activity state template code',
      #      :description => 'Activity state template description',
      #      :subtype => 'IV',
      #      :icon_uri => 'ues:#{system}:#{ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{S10_C02}',
      #      :order => 40,
      #      :entry_state => false,
      #      :exit_state => false,
      #      :participants => ["COMPETENT_ROLE","EXECUTIVE_ROLE","OWNER","OTHERS"],
      #      :digital_workspace_visibility => ["COMPETENT_ROLE","EXECUTIVE_ROLE"])
      def set_attributes(activity_state_template_uri, options)
        svc = UU::OS::REST::RemoteClient.new(ActivityStateTemplate)
        payload = UU::OS::ActivityStateTemplate::ActivityStateTemplateSetAttributes.new(options).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', activity_state_template_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Deletes the specified activity state template. It it not possible to delete the entry or the exit state
      # template, or a state template used in a condition or an action, or if it is important for the system according
      # to its type.
      # 
      # @param [String, UU::OS::UESURI] activity_state_template_uri <b>Not nil.</b> URI of the activity state template which is to be deleted.
      # 
      # @example
      #   # Delete an activity state template
      #   UU::OS::ActivityStateTemplate.delete('ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE')
      def delete(activity_state_template_uri)
        svc = UU::OS::REST::RemoteClient.new(ActivityStateTemplate)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('delete', activity_state_template_uri)
        end
      end

      # Returns a list of activity state templates for the specified activity template according to entered criteria.
      # 
      # @param [String, UU::OS::UESURI] activity_template_uri <b>Not nil.</b> URI of the activity template whose activity state templates are to be listed.
      # @param [UU::OS::ActivityStateTemplate::ActivityStateTemplateGetActivityStateTemplateList] criteria Specific options.
      # @option criteria [String] :query The query for filtering and ordering of the result list. It has to conform the ues query syntax (see documentation for more info). If nil is set, no filtering is applied and entries are ordered by the parameter order and name as the second criteria. Otherwise, it is possible to filter and sort the result list by the following attributes: name, code, type, entryState, exitState and order.
      # 
      # @see https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation.
      # 
      # @return [UU::OS::REST::ResultList<ActivityStateTemplate::ActivityStateTemplateGetActivityStateTemplateList, ActivityStateTemplate::ActivityStateTemplateAttributes>] List of activity state templates according to the entered entity. Contains instances of {UU::OS::ActivityStateTemplate::ActivityStateTemplateAttributes}.
      # 
      # @example
      #   # Return list of activity state templates which are not the exit state and have the initial type
      #   UU::OS::ActivityStateTemplate.get_activity_state_template_list('ues:TERRITORY:META_ARTIFACT:ACTIVITY_TEMPLATE',
      #      :query => 'type="INITIAL" AND exit_state=false')
      def get_activity_state_template_list(activity_template_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(ActivityStateTemplate)
        dto = ActivityStateTemplate::ActivityStateTemplateGetActivityStateTemplateList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getActivityStateTemplateList', activity_template_uri)
          return UU::OS::REST::ResultList.new(ActivityStateTemplate::ActivityStateTemplateGetActivityStateTemplateList, ActivityStateTemplate::ActivityStateTemplateAttributes, res)
        end
      end

    end
  end

end
