require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/app_meta_model/app_meta_model_create'
require 'uu/os/app_meta_model/app_meta_model_attributes'
require 'uu/os/app_meta_model/app_meta_model_set_attributes'
require 'uu/os/app_meta_model/app_meta_model_get_entry_list'
require 'uu/os/app_meta_model/app_meta_model_entry_attributes'
require 'uu/os/app_meta_model/app_meta_model_install_app_parameters'
require 'uu/os/app_meta_model/app_meta_model_install_shared_meta_model_app_parameters'

module UU
  module OS
    module AppMetaModel
      extend self

      # Service path
      PATH = 'uu/os/AppMetaModel'
      # Creates a new app meta model. The command creates a new app meta model in a specified location. At the very least meta
      # artifact must be specified in {AppMetaModelCreate} DTO. Competent role for the app meta model will be selected as
      # the most suitable according to specified container and meta artifact (executive/authorized role with connected interface), or can be
      # also specified in {AppMetaModelCreate} DTO.
      #
      # @param [String, UU::OS::UESURI] location_uri UESURI of the folder/org. unit/meta model where app meta model will be created.
      # @param [AppMetaModelCreate] app_meta_model DTO containing attributes of the new app meta model
      # @option app_meta_model [String, UU::OS::UESURI] :meta_artifact_uri UESURI of the meta artifact. Only a meta artifact with the template class of App Meta Model (not its descendants) can be used. If the rights does not permit to create a app meta model from the meta artifact in the specified location, the create command fails.
      # @option app_meta_model [String] :code Code of the new app meta model. If nil is set, the code is generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails.
      # @option app_meta_model [String] :app_box_uri UESURI of appbox from which was the app meta model created.
      # @option app_meta_model [String] :name Name of the new app meta model. If nil is set, tName from meta artifact is used if set, otherwise the create command fails.
      # @option app_meta_model [String] :description Description of the new app meta model. If nil is set, tDescription from meta artifact is used if set, otherwise no description is set on the new app meta model.
      # @option app_meta_model [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role. If nil is set, the executive role from authorization is used. If no role is authorized to create the new app meta model into specified container, the create command fails. If the selected role doesn't have connected the role interface with the selected meta artifact, the create command fails. If the selected role doesn't belong to one of superior organizational units, the create command fails.
      # @option app_meta_model [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. If nil is set, security level is taken from the meta artifact tSecurityLevel. If the meta artifact tSecurityLevel is higher than theuser's security clearance, the create command fails.
      # @option app_meta_model [String, UU::OS::UESURI] :icon_uri URI of the app meta model's icon. If nil is set, the app meta model uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option app_meta_model [Boolean] :component_content Defines whether the primarily displayed content of the created AppMetaModel is component content. Default value is false.
      #
      # @return [UU::OS::UESURI] UESURI of the new app meta model
      #
      # @example
      #   # Create a new app meta model.
      #   UU::OS::AppMetaModel.create('ues:TERRITORY:ORGANIZATIONAL_UNIT', :code => 'NEW_APP_META_MODEL', :name => 'New App Meta Model')
      def create(location_uri, app_meta_model = nil)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        payload = UU::OS::AppMetaModel::AppMetaModelCreate.new(app_meta_model).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', location_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns attributes representing the app meta model specified by the app_meta_model_uri parameter. The command
      # does not change the state of the app meta model in the system.
      #
      # @param [String, UU::OS::UESURI] app_meta_model_uri UESURI of an app meta model whose attributes are returned
      #
      # @return [AppMetaModelAttributes] DTO containing the app meta model attributes
      #
      # @example
      #   # Get attributes of an app meta model.
      #   UU::OS::AppMetaModel.get_attributes('ues:TERRITORY:APP_META_MODEL')
      def get_attributes(app_meta_model_uri)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', app_meta_model_uri)
          return AppMetaModel::AppMetaModelAttributes.new(res)
        end
      end

      # Command for setting attributes of an app meta model. The command can't modify references to objects like location, competent role
      # and so on.
      #
      # @param [String, UU::OS::UESURI] app_meta_model_uri UESURI of the app meta model whose attributes are to be changed
      # @param [AppMetaModelSetAttributes] app_meta_model DTO containing new attributes the app meta model
      # @option app_meta_model [String] :name The new name of the app meta model. If nil is set, the name is not modified.
      # @option app_meta_model [String] :code The new code of the app meta model. If nil is set, the code is not modified. The code must be unique in a given territory, otherwise the set_attributes command fails.
      # @option app_meta_model [String] :description The new description of the app meta model. If nil is set, the description is not modified.
      # @option app_meta_model [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the app meta model. If nil is set, the security level is not modified. If the meta artifact tSecurityLevel is higher than the user's security clearance, the set_attributes command fails.
      # @option app_meta_model [String, UU::OS::UESURI] :icon_uri URI of the app meta model's icon. If nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the app meta model's icon is removed and the app meta model uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option app_meta_model [Boolean] :component_content Defines whether the primarily displayed content of the created AppMetaModel is component content. Default value is false.
      #
      # @return [UU::OS::UESURI] UESURI of the app meta model
      #
      # @example
      #   # Set a different name and security_level of an existing app meta model.
      #   UU::OS::AppMetaModel.set_attributes('ues:TERRITORY:APP_META_MODEL', :name => 'New Name of App Meta Model', :security_level => UU::OS::Artifact::SecurityLevel::INTERNAL_USE)
      def set_attributes(app_meta_model_uri, app_meta_model = nil)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        payload = UU::OS::AppMetaModel::AppMetaModelSetAttributes.new(app_meta_model).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', app_meta_model_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # This command returns a list of entries (artifacts or shortcuts) enlisted in the specified app meta model. Only entries listed
      # exactly in the specified app meta models are considered (so subordinate app meta models are not considered). The command does not
      # return all entries, only a sublist. The list of returned entries could be filtered by the name, code or type of an entry. The list is
      # sorted by the name (and codes in case are equal) of a entry by default, or it can be sorted arbitrarily by any of entry attributes
      # usable for the filtering or these attributes combinations. See the UES query documentation for more information.
      #
      # See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      #
      # @param [String, UU::OS::UESURI] app_meta_model_uri UESURI of the app meta model from where entries (artifacts or shortcuts) are listed.
      # @param [AppMetaModelGetEntryList] criteria DTO containing criteria for listing entries. Available parameters are: {AppMetaModelEntryAttributes#name name}, {AppMetaModelEntryAttributes#code code} and {AppMetaModelEntryAttributes#type type}.
      # @option criteria [String] :query The <b>UES query</b> string. It has to conform the ues query syntax (see documentation for more info). Filtering and order of the result list can be specified there. If nil is set, no filtering is applied and entries are ordered by name of referenced artifact (and code and type as the second and the third criteria).
      # @option criteria [Boolean] :recursive Flag for recursive listing. It means that if the listed app meta model contains inner folders and the recursive flag is true, entries of the all inner folders are in result too. If the recursive flag is nil or false, only entries are returned exactly from the listed app meta model.
      #
      # @return [UU::OS::REST::ResultList<AppMetaModel::AppMetaModelGetEntryList, AppMetaModel::AppMetaModelEntryAttributes>] List of DTO representing entries of the app_meta_model_uri (artifacts or shortcuts)
      #
      # @example
      #   # Returns list of entries, whose names start with 'T', from an app meta model.
      #   UU::OS::AppMetaModel.get_entry_list('ues:TERRITORY:APP_META_MODEL',
      #      :query => 'name="T%"')
      def get_entry_list(app_meta_model_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        dto = AppMetaModel::AppMetaModelGetEntryList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        svc.add_parameter('recursive', dto.recursive)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getEntryList', app_meta_model_uri)
          return UU::OS::REST::ResultList.new(AppMetaModel::AppMetaModelGetEntryList, AppMetaModel::AppMetaModelEntryAttributes, res)
        end
      end
      
      # Command UU::OS::AppMetaModel.install is used to install an application to the Business or Development territory. It also allows to install applications with shared meta model, if the application supports it. In this case, first will be performed the installation to the system territory (or to the location, specified in the parameter devSharedSpaceUri, when installed to Development territory) and then the application will be installed in specified Business territory.
      #
      # @param [String, UU::OS::UESURI] territory_uri UESURI of territory, in which the application should be installed.
      # @param [AppMetaModelInstallAppParameters] app_parameters DTO containing attributes of application.
      # @option app_parameters [String] :app_box_uri UESURI of uuApp Box, according to which the application should be installed. The uuEE of the territory, in which the application is going to be installed, must have rights on copying of mentioned uuApp Box.  
      # @option app_parameters [Hash] :app_limits JSON, which defines operating limits of the application.
      # @option app_parameters [Array<Hash>] :app_identities The parameter is used for entering the identities, needed for installation of the application. Obligatorily must be entered the identities "appInstaller" and "appIdentity".
      #   * :identity (String) - Identifier of the identity - "appIdentity", "appInstaller".
      #   * :uid (String) - UID of the uuEE (identity), which should represent the account.
      #   * :qos (String) - JSON, which defines Local and Remote QoS.
      #   * :credentials (Array<Hash>) - Array, which contains identity credentials. The parameter is set only for identity "appInstaller".
      #     * :password_file_path (String) - Path to a file on server, which defines the encrypted login data of application identity. Relative address only.
      #     * :access_code1 (String) - Access code 1 of application identity.
      #     * :access_code2 (String) - Access code 2 of application identity.
      # @option app_parameters [Hash] :app_installation_config JSON with additional configuration of application, which can be necessary during its installation.
      # @option app_parameters [Array<Hash>] :app_identity_cast_list JSON, with the list of casts of identities "appIdentity" and "appInstaller".
      #   * :app_identity_to_cast (String) - Identity to cast - "appIdentity" or "appInstaller".
      #   * :subject_code (String) - Code of a role or a group, to which the identity "appIdentity" or "appInstaller" will be cast. The identity is always cast to EXECUTIVE level.
      # @option app_parameters [Boolean] :force Parameter is used for forcing the installation of uuApp in case if the metamodel state of the same uuApp version, which is already installed, is in Problem state. Default value is false.
      # @option app_parameters [String, UU::OS::UESURI] :dev_system_space_uri Parameter is used for installation in development mode. Defines UESURI of the location in the Business space of the territory, where the uuApp Meta Model is going to be installed.
      # @option app_parameters [String, UU::OS::UESURI] :dev_app_space_uri Parameter is used for installation in development mode. Defines UESURI of the location in the Business space of the territory, where the uuApp structure is going to be installed.
      # @option app_parameters [String] :dev_meta_model_suffix Parameter is used for installation in development mode. In case of installation of a new application, it is needed to add a suffix, which will be added to the artifact codes of the application.
      # @option app_parameters [String, UU::OS::UESURI] :dev_shared_space_uri Parameter is used for installation in development mode. Defines UESURI of the location, where the shared uuApp Meta Model is going to be installed.
      #
      # @return [UU::OS::REST::Future] Instance of future
      #
      # @example
      #   # Returns UU::OS::REST::Future
      #   UU::OS::AppMetaModel.install('ues:UU-BT:UU-BT',
      #       :app_box_uri => 'ues:OP-BT:WHS.CDS/CDS.GUESTBOOK_1.0.0',
      #       :app_identities => [
      #         {
      #           :identity => 'appInstaller',
      #           :uid => '33-6677-1',
      #           :credentials => {
      #             "password_file_path" => 'PASSWORD_FILE'
      #           }
      #         },
      #         {
      #           :identity => 'appIdentity',
      #           :uid => '23-5555-1'
      #         }
      #       ],
      #       :app_identity_cast_list => [
      #         {
      #           :app_identity_to_cast => "appIdentity",
      #           :subject_code => "APPS~AUTHORITIES"
      #         }
      #       ]
      #   )
      #
      #   UU::OS::AppMetaModel.install('ues:UU-BT:UU-BT',
      #       :app_box_uri => 'ues:OP-BT:WHS.CDS/CDS.GUESTBOOK_1.0.0',
      #       :app_identities => [
      #         {
      #           :identity => 'appInstaller',
      #           :uid => '33-6677-1',
      #           :credentials => {
      #             "access_code_1" => "123",
      #             "access_code_2" => "456"
      #           }
      #         },
      #         {
      #           :identity => 'appIdentity',
      #           :uid => '23-5555-1'
      #         }
      #       ],
      #       :dev_system_space_uri => "ues:DEV-BT:SYS",
      #       :dev_app_space_uri => "ues:DEV-BT:APPS",
      #       :dev_shared_space_uri => "ues:DEV-BT:SYS"
      #   )
      def install(territory_uri, app_parameters = nil)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        payload = UU::OS::AppMetaModel::AppMetaModelInstallAppParameters.new(app_parameters).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('install', territory_uri, payload)
          uesuri =  UU::OS::UESURI.new(res)
          return UU::OS::REST::Future.new(uesuri)
        end
      end
      
      # Command UU::OS::AppMetaModel.installSharedMetamodel is used only to install a shared meta model to the system and it can not be used for installation to business territory. Parameters devSystemSpaceUri and devAppSpaceUri are not used within this command.
      #
      # @param [String, UU::OS::UESURI] territory_uri UESURI of territory, in which the application should be installed.
      # @param [AppMetaModelInstallSharedMetaModelAppParameters] app_parameters DTO containing attributes of application.
      # @option app_parameters [String] :app_box_uri UESURI of uuApp Box, according to which the application should be installed. The uuEE of the territory, in which the application is going to be installed, must have rights on copying of mentioned uuApp Box.  
      # @option app_parameters [Hash] :app_limits JSON, which defines operating limits of the application.
      # @option app_parameters [Array<Hash>] :app_identities Parameter is used for entering the identities, needed for installation of the application. Obligatorily must be entered the identity "appInstaller".
      #   * :identity (String) - Identifier of the identity - "appIdentity", "appInstaller".
      #   * :uid (String) - UID of the uuEE (identity), which should represent the account.
      #   * :qos (String) - JSON, which defines Local and Remote QoS.
      #   * :credentials (Array<Hash>) - Array, which contains identity credentials. The parameter is set only for identity "appInstaller".
      #     * :password_file_path (String) - Path to a file on server, which defines the encrypted login data of application identity. Relative address only.
      #     * :access_code1 (String) - Access code 1 of application identity.
      #     * :access_code2 (String) - Access code 2 of application identity.
      # @option app_parameters [Hash] :app_installation_config JSON with additional configuration of application, which can be necessary during its installation.
      # @option app_parameters [Array<Hash>] :app_identity_cast_list JSON, with the list of casts of identities "appIdentity" and "appInstaller".
      #   * :app_identity_to_cast (String) - Identity to cast - "appIdentity" or "appInstaller".
      #   * :subject_code (String) - Code of a role or a group, to which the identity "appIdentity" or "appInstaller" will be cast. The identity is always cast to EXECUTIVE level.
      # @option app_parameters [Boolean] :force Parameter is used for forcing the installation of uuApp in case if the metamodel state of the same uuApp version, which is already installed, is in Problem state. Default value is false.
      # @option app_parameters [String] :dev_meta_model_suffix Parameter is used for installation in development mode. In case of installation of a new application, it is needed to add a suffix, which will be added to the artifact codes of the application.
      # @option app_parameters [String, UU::OS::UESURI] :dev_shared_space_uri Parameter is used for installation in development mode. Defines UESURI of the location, where the shared uuApp Meta Model is going to be installed.
      #
      # @return [UU::OS::REST::Future] Instance of future
      #
      # @example
      #   # Returns UU::OS::REST::Future
      #   UU::OS::AppMetaModel.install_shared_meta_model('ues:SYSTER-BT:SYSTER-BT',
      #       :app_box_uri => 'ues:OP-BT:WHS.CDS/CDS.GUESTBOOK_1.0.0',
      #       :app_identities => [
      #         {
      #           :identity => 'appInstaller',
      #           :uid => '33-6677-1',
      #           :credentials => {
      #             "password_file_path" => 'PASSWORD_FILE'
      #           }
      #         }
      #       ]
      #   )
      def install_shared_meta_model(territory_uri, app_parameters = nil)
        svc = UU::OS::REST::RemoteClient.new(AppMetaModel)
        payload = UU::OS::AppMetaModel::AppMetaModelInstallSharedMetaModelAppParameters.new(app_parameters).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('installSharedMetaModel', territory_uri, payload)
          uesuri =  UU::OS::UESURI.new(res)
          return UU::OS::REST::Future.new(uesuri)
        end
      end

    end
  end
end
