require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/app_package/app_package_create'
require 'uu/os/app_package/app_package_attributes'
require 'uu/os/app_package/app_package_entry_attributes'
require 'uu/os/app_package/app_package_export'
require 'uu/os/app_package/app_package_get_export_data'
require 'uu/os/app_package/app_package_set_attributes'
require 'uu/os/app_package/app_package_get_entry_list'

module UU
  module OS
    module AppPackage
      extend self

      # Service path
      PATH = 'uu/os/AppPackage'

      # Creates a new app package. The command creates a new app package in a specified location.
      # At the very least meta artifact and name must be specified in {AppPackageCreate}
      # Object. Competent role for the app package will be selected as the most suitable according
      # to specified container and meta artifact (executive/authorized role with connected
      # interface), or can be also specified in {AppPackageCreate} Object. 
      #
      # @param [String, UU::OS::UESURI] location_uri UESURI of the folder/org. unit/meta model where app package will be created.
      # @param [AppPackageCreate] app_package Object containing attributes of the new app package. 
      # @option app_package [String, UU::OS::UESURI] :meta_artifact_uri UESURI of the meta artifact. Only a meta artifact with the template class of meta model (not its descendants) can be used. If the rights does not permit to create a new app package from the meta artifact in the specified location, the create command fails.
      # @option app_package [String] :name Name of the new app package. If nil is set, template name from meta artifact is used if set, otherwise the create command fails.
      # @option app_package [String] :code Code of the new app package. If nil is set, the code is generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails.
      # @option app_package [String] :description Description of the new app package. If nil is set, from meta artifact is used if set, otherwise no description is set on the new app package.
      # @option app_package [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role. If nil is set, the executive role from authorization is used. If no role is authorized to create the new app package into specified container, the create command fails. If the selected role doesn't have connected the role interface with the selected meta artifact, the create command fails.
      # @option app_package [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. If nil is set, security level is taken from the meta artifact default security level. If the meta artifact template security level is higher than the user's security clearance, the create command fails.
      # @option app_package [String, UU::OS::UESURI] :icon_uri URI of the app package's icon. If nil is set, the app package uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      #
      # @return [UU::OS::UESURI] UESURI of the new app package.
      #       
      # @example
      #   # Create a new app package.
      #   UU::OS::AppPackage.create('ues:TERRITORY:ORGANIZATIONAL_UNIT', :code => 'NEW_APP_PACKAGE', :name => 'New App Package')
      def create(location_uri, app_package = nil)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        payload = UU::OS::AppPackage::AppPackageCreate.new(app_package).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', location_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns attributes representing the app package specified by the
      # app_package_uri parameter. The command does not change the state of the app package
      # in the system.
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of an app package whose attributes are returned
      #
      # @return [AppPackageAttributes] Object containing the app package attributes
      # 
      # @example
      #   # Get attributes of the specified app package.
      #   attributes = UU::OS::AppPackage.get_attributes('ues:TERRITORY:APP_PACKAGE')
      def get_attributes(app_package_uri)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', app_package_uri)
          return AppPackage::AppPackageAttributes.new(res)
        end
      end

      # Command for setting attributes of an app package. The command can't modify references
      # to objects like location, competent role and so on.
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of the app package whose attributes are to be changed
      # @param [AppPackageSetAttributes] app_package Object containing new attributes the app package
      # @option app_package [String] :name The new name of the app package. If nil is set, the name is not modified.
      # @option app_package [String] :code The new code of the app package. If nil is set, the code is not modified. When blank string is set, the code is generated automatically using the system sequence. The code must be unique in a given territory, otherwise the set_attributes command fails.
      # @option app_package [String] :description The new description of the app package. If nil is set, the description is not modified.
      # @option app_package [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the app package. If nil is set, the security level is not modified. If the app package dictionary security level is higher than the user's security clearance, the set_attributes command fails.
      # @option app_package [String, UU::OS::UESURI] :icon_uri URI of the app package's icon. When nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the app package's icon is removed and the app package uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      #
      # @return [UU::OS::UESURI] UESURI of the app package
      # 
      # @example
      #   # Set basic attributes of an app package.
      #   UU::OS::AppPackage.set_attributes('ues:TERRITORY:APP_PACKAGE', :code=>'NEW_CODE', :name=>'App Package name', :description=>'Some sample description.')
      def set_attributes(app_package_uri, app_package = nil)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        payload = UU::OS::AppPackage::AppPackageSetAttributes.new(app_package).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', app_package_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Deletes the specified app package from the system. The app package is
      # deleted even if it contains a locked sheet or attachment. The app package cannot
      # be deleted when:
      #   the app package is not empty
      #   the app package's workflow is not created
      #   a property contains a reference to this app package
      #   a time sheet exists with a reference to this app package
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of the app package which is about to be deleted.
      # 
      # @example
      #   # Delete an app package with specified uri.
      #   UU::OS::AppPackage.delete('ues:TERRITORY:APP_PACKAGE')
      def delete(app_package_uri)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', app_package_uri)
        end
      end

      # Returns a list of entries (artifacts, meta artifacts, meta interfaces,
      # containers, app package dictionaries or shortcuts) enlisted in the specified
      # app package dictionary. Command does not return all entries, only a sublist. The
      # list of returned entries could be filtered by the name, code or entityTypeUri of
      # an entry. The list is sorted by the name (and codes in case are equal) of a entry by
      # default, or it can be sorted arbitrarily by any of entry attributes usable for the
      # filtering or these attributes combinations.
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of the app package dictionary from where entries are listed.
      # @param [AppPackageGetEntryList] criteria Object containing criteria. Available parameters are: {AppPackageEntryAttributes#name name}, {AppPackageEntryAttributes#code code} and {AppPackageEntryAttributes#entityTypeUri entityTypeUri}. See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      # @option criteria [String] :query The <b>UES query</b> string. It has to conform the ues query syntax (see documentation for more info). Filtering and order of the result list can be specified there. If nil is set, no filtering is applied and entries are ordered by name and code as the second criteria.
      # @option criteria [Boolean] :recursive Flag for recursive listing. It means that if the listed folder contains inner folders and the recursive flag is true, entries of the all inner folders are in result too. When the recursive flag is nil or false, only entries are returned exactly from the listed app package.
      #
      # @return [UU::OS::REST::ResultList<AppPackage::AppPackageGetEntryList, AppPackage::AppPackageEntryAttributes>] List of Objects representing app package dictionary entries.
      # 
      # @example
      #   # Get entries of an app package in descending order and even with entries of inner app packages - flag recursive is set to true.
      #   entryList = UU::OS::AppPackage.get_entry_list('ues:TERRITORY:APP_PACKAGE', :query=>'ORDER BY name DESC', :recursive=>true)
      def get_entry_list(app_package_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        dto = AppPackage::AppPackageGetEntryList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        svc.add_parameter('recursive', dto.recursive)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getEntryList', app_package_uri)
          return UU::OS::REST::ResultList.new(AppPackage::AppPackageGetEntryList, AppPackage::AppPackageEntryAttributes, res)
        end
      end

      # Exports an app package to the XML file, which is saved to the export storage. The export is an asynchronous process. <br>
      # It is possible to wait for the end of the process. More information in {UU::OS::Env::Process Process} and in {UU::OS::REST::Future Future}.
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of the app package to export.
      #
      # @return [UU::OS::UESURI] UESURI of the export process.
      #
      # @example
      #   # Start exporting an app package.
      #   process_uri = UU::OS::AppPackage.export('ues:TERRITORY:APP_PACKAGE')
      #
      #   # Wait for the process to finish.
      #   export_uri = UU::OS::REST::Future.new(proces).get(:timeout=>600)
      def export(app_package_uri)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        payload = UU::OS::AppPackage::AppPackageExport.new(nil).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('export', app_package_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a XML file generated by the {AppPackage#export export} command.
      #
      # @param [String, UU::OS::UESURI] app_package_uri UESURI of the app package we want to get the exported XML file from.
      # @param [UU::OS::AppPackage::AppPackageGetExportData] app_package Specific options.
      # @option app_package [String, UU::OS::Artifact::ExportMimeType] :requested_mime_type MIME type of the returned value. ZIP is returned by default.
      #
      # @return [UU::OS::REST::BinaryValue] Data file of an exported app package (XML or ZIP containing XML).
      #
      # @example
      #   # Start exporting an app package
      #   process_uri = UU::OS::AppPackage.export('ues:TERRITORY:APP_PACKAGE')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::AppPackage.get_export_data(export_uri, :requested_mime_type => 'application/xml')
      def get_export_data(app_package_uri, app_package = nil)
        svc = UU::OS::REST::RemoteClient.new(AppPackage)
        dto = UU::OS::AppPackage::AppPackageGetExportData.new(app_package)
        if dto.requested_mime_type
          svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        end
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getExportData', app_package_uri)
        end
      end

    end
  end

end
