require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/meta_group/meta_group_add_role_interface'
require 'uu/os/meta_group/meta_group_attributes'
require 'uu/os/meta_group/meta_group_copy'
require 'uu/os/meta_group/meta_group_create'
require 'uu/os/meta_group/meta_group_export'
require 'uu/os/meta_group/meta_group_get_export_data'
require 'uu/os/meta_group/meta_group_get_role_interface_list_item'
require 'uu/os/meta_group/meta_group_get_role_interface_list'
require 'uu/os/meta_group/meta_group_remove_role_interface'
require 'uu/os/meta_group/meta_group_set_attributes'

# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Meta Group.
    module MetaGroup
      extend self

      # Service path
      PATH = 'ues/core/role/UESMetaGroup'

      # Returns attributes representing the meta group specified by the meta_group_uri parameter.
      # The command does not change the state of the meta group in the system.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of a meta group whose attributes are returned
      #
      # @return [UU::OS::MetaGroup::MetaGroupAttributes] DTO containing the meta group attributes
      #
      # @example
      #   # Get attributes of a meta group
      #   attributes = UU::OS::MetaGroup.get_attributes('ues:TERRITORY:META_GROUP')
      def get_attributes(meta_group_uri)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', meta_group_uri)
          return MetaGroup::MetaGroupAttributes.new(res)
        end
      end

      # Command for setting attributes of a meta group. The command cannot modify references to objects like
      # location, competent role and so on.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of a meta group whose attributes are returned
      # @param [UU::OS::MetaGroup::MetaGroupSetAttributes] meta_group DTO containing new attributes of the meta group
      # @option meta_group [String] :name The new name of the meta group. If nil is set, the name is not modified.
      # @option meta_group [String] :code The new code of the meta group. If nil is set, the code is not modified. The code must be unique in a given territory.
      # @option meta_group [String] :description The new description of the meta group. If nil is set, the description is not modified.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the meta group. If nil is set, the security level is not modified.
      # @option meta_group [String, UU::OS::UESURI] :icon_uri UESURI of the new meta group's icon. If nil is set, the icon is not modified. If nil UESURI ("ues:{[-1]}:{[-1]}:") is set, the meta group icon is removed and the meta group uses icon from it's source meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_group [String] :default_name Specifies new default name of the created role according to meta group. If nil is set, the default name is not modified.
      # @option meta_group [String] :default_code Specifies new default code of the created role according to meta group. If nil is set, the default code is not modified.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies new default security level of the created role according to meta group. If nil is set, the default security is not modified.
      # @option meta_group [String, UU::OS::UESURI] :default_icon_uri Specifies new default icon of the created role according to meta group (specified in meta_group_uri). If nil is set, the default icon is not modified. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_group [Boolean] :default_searchable_in_territory Specifies if it is possible to search for the role created according to the meta group in its territory. If nil is set, the attribute is not modified.
      # @option meta_group [Boolean] :component_content Defines whether the primarily displayed content of the created Meta group is component content. Default value is false.           
      # @option meta_group [Boolean] :default_component_content Defines whether the primarily displayed content of a group created according to the meta group (specified in source_meta_artifact_uri) is component content. When nil is set, the default component content setting is not modified.
      #
      # @return [UU::OS::UESURI] UESURI of the meta group
      #
      # @example
      #   # Set attributes of a meta role
      #   UU::OS::MetaGroup.set_attributes('ues:TERRITORY:META_GROUP', :name => 'New Meta Group', :code => 'NEW_META_GROUP',
      #     :description => 'Meta Group designated for testing purposes', :default_searchable_in_territory => true)
      def set_attributes(meta_group_uri, meta_group = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        payload = UU::OS::MetaGroup::MetaGroupSetAttributes.new(meta_group).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', meta_group_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Copy a meta group. The command creates a new meta group according to specified source
      # meta group in a specified location. At the very least location meta group and name must
      # be specified in {MetaGroupCopy} DTO. Competent role for the new meta group copy will
      # be selected as the most suitable according to specified container and meta group
      # (executive/authorized role with connected interface), or can be also specified in {MetaGroupCopy} DTO.
      #
      # @param [String, UU::OS::UESURI] source_meta_artifact_uri UESURI of the meta group which will be copied.
      # @param [UU::OS::MetaGroup::MetaGroupCopy] meta_group DTO containing attributes of the new meta group copy.
      # @option meta_group [String, UU::OS::UESURI] :location_uri UESURI of the meta model dictionary where to the meta group will be copied.
      # @option meta_group [String] :name Name of the new meta group copy. The value must not be blank.
      # @option meta_group [String] :code Code of the new meta group copy.
      # @option meta_group [String] :description Description of the new meta group copy.
      # @option meta_group [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :security_level Security level.
      # @option meta_group [String, UU::OS::UESURI] :icon_uri URI of the new meta group's icon. Allowed format of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option meta_group [Boolean] :copy_sheets Specifies whether to copy content from the source meta group to the new copy. Default value is false.
      # @option meta_group [Boolean] :copy_lifecycle Specifies whether to copy workflow from the source meta group to the new copy. Default value is false.
      # @option meta_group [Boolean] :copy_attachments Specifies whether to copy attachments from the source meta group to the new copy. Default value is false.
      # @option meta_group [Boolean] :copy_sheet_templates Specifies whether to copy sheet templates from the source meta group to the new copy. Default value is false.
      # @option meta_group [String] :default_name Specifies default name of the created role according to new meta group copy.
      # @option meta_group [String] :default_code Specifies default code of the created role according to new meta group copy.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies default security level of the created role according to new meta group copy.
      # @option meta_group [String, UU::OS::UESURI] :default_icon_uri Specifies default icon of the created role according to new meta group copy. Allowed format of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option meta_group [String] :default_searchable_in_territory Specifies if it is possible to search for the role created according to the meta group in its territory.
      # @option meta_group [Boolean] :component_content Defines whether the primarily displayed content of the created Meta group is component content. Default value is false.           
      # @option meta_group [Boolean] :default_component_content Defines whether the primarily displayed content of a role created according to the new meta role copy is component content. When nil is set, the default component content setting is copied from the source meta artifact.
      #
      # @return [UU::OS::UESURI] UESURI of the copied meta group.
      #
      # @example
      #   # Copy a meta group (including its sheets); change a few attributes
      #   UU::OS::MetaGroup.copy('ues:TERRITORY:META_GROUP', :location_uri => 'ues:TERRITORY:APP_PACKAGE', :code => 'META_GROUP_COPY',
      #     :icon_uri => 'ues:#{system}:#{ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{ART_077}',
      #     :security_level => UU::OS::Artifact::SecurityLevel::INTERNAL_USE, :copy_sheets => true)
      def copy(source_meta_artifact_uri, meta_group = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        svc.timeout = 120 # may take longer than default 60 seconds
        payload = UU::OS::MetaGroup::MetaGroupCopy.new(meta_group).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('copy', source_meta_artifact_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Creates new role interface connection to the specified meta group.
      # The role interface has to be in an active state and must not be already added. The role
      # interface can be added to meta group only provided that role interface is located in the
      # same, child or parent meta model as meta group.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of the role where the role interface will be connected
      # @param [UU::OS::MetaGroup::MetaGroupAddRoleInterface] role_interface DTO containing role interface that is to be connected
      # @option role_interface [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that is to be connected to the specified meta group.
      #
      # @example
      #   # Connect another role interface to a meta group
      #   UU::OS::MetaGroup.add_role_interface('ues:TERRITORY:META_GROUP', :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE')
      def add_role_interface(meta_group_uri, role_interface = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        payload = UU::OS::MetaGroup::MetaGroupAddRoleInterface.new(role_interface).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('addRoleInterface', meta_group_uri, payload)
        end
      end

      # Returns list of the role interfaces that are added to the specified meta group.
      # The list of returned interfaces could be filtered by the name or code. Returned list does
      # not contain any objects filtered out by executed UESQuery. The list is sorted by the name
      # (and codes in case are equal) of an interface by default.
      #
      # See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of the meta group for which the added role interfaces are to be listed
      # @param [UU::OS::MetaGroup::MetaGroupGetRoleInterfaceList] criteria DTO containing criteria for listing connected role interfaces
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {uuQuery - Guideline}[https://plus4u.net/ues/sesm?SessFree=ues%253AUNI-BT%255B210795%255D%253AUES.DOC%252FCMD_V1%252FUESQUERY%255B44191574519623267%255D%253A]). It is possible to filter the result list by the name or code.
      #
      # @return [UU::OS::REST::ResultList<MetaGroup::MetaGroupGetRoleInterfaceList, MetaGroup::MetaGroupGetRoleInterfaceListItem>] List of DTO representing role interfaces which are connected to the specified meta group
      #
      # @example
      #   # Get a list of all role interfaces which are connected to the specified meta group
      #   UU::OS::MetaGroup.get_role_interface_list('ues:TERRITORY:META_GROUP')
      #
      #   # Get a list of role interfaces with name that starts with string John.
      #   UU::OS::MetaGroup.get_role_interface_list('ues:TERRITORY:META_GROUP', :query => "name='John%'")
      def get_role_interface_list(meta_group_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        dto = MetaGroup::MetaGroupGetRoleInterfaceList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getRoleInterfaceList', meta_group_uri)
          return UU::OS::REST::ResultList.new(MetaGroup::MetaGroupGetRoleInterfaceList, MetaGroup::MetaGroupGetRoleInterfaceListItem, res)
        end
      end

      # The command removes role interface connection from the specified meta group.
      # Only added interfaces can be removed from meta group.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of the meta group where the role interface will be disconnected.
      # @param [UU::OS::MetaGroup::MetaGroupRemoveRoleInterface] role_interface DTO containing role interface that to be disconnected
      # @option role_interface [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that to be removed from the specified meta group.
      #
      # @example
      #   # Remove a connected role interface from a meta group
      #   UU::OS::MetaGroup.remove_role_interface('ues:TERRITORY:META_GROUP', :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE')
      def remove_role_interface(meta_group_uri, role_interface = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        payload = UU::OS::MetaGroup::MetaGroupRemoveRoleInterface.new(role_interface).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('removeRoleInterface', meta_group_uri, payload)
        end
      end

      # Exports a meta group to the XML file, which is saved to the export storage. The export is an asynchronous process.<br>
      # It is possible to wait for the end of the process. More information in {UU::OS::Env::Process Process} and in {UU::OS::REST::Future Future}.<br>
      # To get exported data from the storage use {#get_export_data get_export_data}.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of the meta group to export.
      #
      # @return [UU::OS::UESURI] UESURI of the export process.
      #
      # @example
      #   # Start exporting a meta group.
      #   process_uri = UU::OS::MetaGroup.export('ues:TERRITORY:META_GROUP')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::MetaGroup.get_export_data(export_uri)
      def export(meta_group_uri)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        payload = UU::OS::MetaGroup::MetaGroupExport.new(nil).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('export', meta_group_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a XML file generated by the export command.
      #
      # @param [String, UU::OS::UESURI] meta_group_uri UESURI of the meta artifact we want to get the exported XML file from.
      # @param [UU::OS::MetaGroup::MetaGroupGetExportData] meta_group Specific options.
      # @option meta_group [String] :requested_mime_type MIME type of the returned value. Possible values are 'application/xml' for XML and 'application/zip' for ZIP. ZIP is returned by default.
      #
      # @return [UU::OS::REST::BinaryValue] Data file of an exported meta artifact (XML or ZIP containing XML).
      #
      # @example
      #   # Start exporting a meta group.
      #   process_uri = UU::OS::MetaGroup.export('ues:TERRITORY:META_GROUP')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::MetaGroup.get_export_data(export_uri)
      def get_export_data(meta_group_uri, meta_group = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        dto = UU::OS::MetaGroup::MetaGroupGetExportData.new(meta_group)
        if dto.requested_mime_type
          svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        end
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getExportData', meta_group_uri)
        end
      end

      # Creates a meta group from an exported XML file in the specified location. Source meta group must be derived (at least indirectly) from a meta group which is a part of UU.OS uuApp. Creation of meta group is divided into synchronous and asynchronous part. During the synchronous part, the meta group is created. During the asynchronous part, the template objects are created. If the asynchronous process fails and the Problem state is set on meta group, it is possible to solve the problem (e.g. add reference_map) and run the command again.
      # @deprecated The uuCommand MetaGroup.create is used for creating meta group from the XML file temporarily. This feature will be removed in a future version and replaced by the specific uuCommands for creating template objects.
      #
      # @param [String, UU::OS::UESURI] location_uri UESURI of the meta model dictionary where new meta group should be created.
      # @param [UU::OS::MetaGroup::MetaGroupCreate] meta_group Specific options.
      # @option meta_group [String, File, UU::OS::REST::BinaryValue] :data XML or zip compressed XML containing informations about meta group (e.g. exported with {export})
      # @option meta_group [String, UU::OS::UESURI] :competent_role_uri UESURI of the role that will be competent for the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String] :name Name of the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String] :code Code of the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String] :description Description of the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :security_level Security level of the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String, UU::OS::UESURI] :icon_uri UESURI of the new meta group icon. If nil is set, the attribute from the XML (entered in :data option) is used. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_group [String] :default_name Default name of a group that will be created according to the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String] :default_code Default code of a group that will be created according to the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [String, UU::OS::UESURI] :default_icon_uri UESURI of the artifact icon that will be created according to the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_group [String, UU::OS::Artifact::SecurityLevel] :default_security_level Default security level of the new meta group. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_group [Boolean] :default_searchable_in_territory Flag indicating whether a role that will be created according to the new meta group is searchable in the territory.
      # @option meta_group [Boolean] :import_life_cycle_template Flag indicating whether the life cycle template will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_property_templates Flag indicating whether the property templates will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_sheet_templates Flag indicating whether the sheet templates will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_properties Flag indicating whether the properties will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_sheets Flag indicating whether the sheets will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_attachments Flag indicating whether the attachments will be created according to specification in the entered XML. Default value is true.      
      # @option meta_group [Boolean] :import_meta_permissions Flag indicating whether the meta permissions will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_permissions Flag indicating whether the permissions will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_use_cases Flag indicating whether the use cases will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [Boolean] :import_masks Flag indicating whether the permission masks will be created according to specification in the entered XML. Default value is true.
      # @option meta_group [String] :reference_map JSON describing the linkage of references specified in XML to another objects (e.g. references in properties, activity related artifacts. ...).
      # @option meta_group [Boolean] :component_content Defines whether the primarily displayed content of the created Meta group is component content. Default value is false.           
      # @option meta_group [Boolean] :default_component_content Defines whether the primarily displayed content of a group created according to the new meta group is component content. When nil is set, the attribute from the XML (entered in :data option) is used.
      #
      # @return [UU::OS::UESURI] UESURI of the new meta group.
      #
      # @example
      #   # Get exported data from an existing meta group.
      #   exported_data = UU::OS::MetaGroup.get_export_data('ues:TERRITORY:META_GROUP', :requested_mime_type => 'application/xml')
      #
      #   # Create a new meta group according to exported one without creating sheets and attachments specified in the XML. Also set different code and name.
      #   UU::OS::MetaGroup.create('ues:TERRITORY:APP_PACKAGE', :code => 'NEW_META_GROUP', :name => 'New Meta Group', :data => exported_data.data, :import_sheets => false, :import_attachments => false)
      #
      #   # Create a new meta group and set new objects to the references.
      #   UU::OS::MetaGroup.create('ues:TERRITORY:APP_PACKAGE', :data => exported_data.data, :reference_map => '[{"ues:OLD_TERRITORY:OLD_ARTIFACT1":"ues:TERRITORY:ARTIFACT1"},{"ues:OLD_TERRITORY:OLD_ARTIFACT2:OLD_OBJECT":"ues:TERRITORY:ARTIFACT2:OBJECT"}]')
      def create(location_uri, meta_group = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaGroup)
        dto = UU::OS::MetaGroup::MetaGroupCreate.new(meta_group)
        payload = dto.to_hash
        payload.delete(:data)
        tmpfile = Tempfile.new("MetaGroupCreate")
        begin
          if (!dto.data.data.respond_to?(:read))
            tmpfile.write(dto.data.data)
            payload[:data] = tmpfile
          elsif (dto.data.data.respond_to?(:read) && !dto.data.data.respond_to?(:path))
            tmpfile.binmode
            dto.data.data.rewind
            tmpfile.write(dto.data.data.read)
            payload[:data] = tmpfile
          else
            payload[:data] = dto.data.to_hash(false)
          end
          tmpfile.flush
          tmpfile.rewind
              
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', location_uri, payload)
            return UU::OS::UESURI.new(res)
          end
        ensure
          tmpfile.close
          tmpfile.unlink
        end
      end

    end
  end

end
